package laboratoire4;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

public class MiniMax {
    private static final int MAX_DEPTH = 4;

    public static MiniMaxResult miniMax(PusherBoard board) {
        long startMillis = System.currentTimeMillis();
        MovingBoard game = new MovingBoard(board.getBoard(), board.getPlayer());
        Collection<Action> actions = getActions(game, true);

        MiniMaxResult maxResult = null;

        for (Action action : actions) {
            int score = min(game, 0, Integer.MIN_VALUE, Integer.MAX_VALUE);

            if (maxResult == null || score > maxResult.getScore()) {
                MovingPawn pawn = action.getPawn();
                maxResult = new MiniMaxResult(score, pawn.getRow(), pawn.getCol(), action.getMovement());
            }
        }

        long endMillis = System.currentTimeMillis();
        System.out.printf("%d ms\n", (endMillis - startMillis));

        return maxResult;
    }

    private static int max(MovingBoard game, int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate(game);
        }

        int maxScore = Integer.MIN_VALUE;

        for (Action action : getActions(game, true)) {
            MovingPawn pawn = action.getPawn();
            Pawn.PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = min(game, depth + 1, Math.max(alpha, maxScore), beta);
            game.revertMove();

            if (score > maxScore) {
                maxScore = score;
            }

            if (maxScore >= beta) {
                return maxScore;
            }
        }

        return maxScore;
    }

    private static int min(MovingBoard game, int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluate(game);
        }

        int minScore = Integer.MAX_VALUE;

        for (Action action : getActions(game, false)) {
            MovingPawn pawn = action.getPawn();
            Pawn.PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = max(game, depth + 1, alpha, Math.min(beta, minScore)) * -1;
            game.revertMove();

            if (score < minScore) {
                minScore = score;
            }

            if (minScore <= alpha) {
                return minScore;
            }
        }

        return minScore;
    }

    private static Collection<Action> getActions(MovingBoard game, boolean max) {
        List<Action> actions = new ArrayList<>();
        Collection<MovingPawn> pawns = max ? game.getMaxPawns() : game.getMinPawns();
        Pawn.PawnMovement[] movements = Pawn.PawnMovement.values();

        for (MovingPawn pawn : pawns) {
            for (Pawn.PawnMovement movement : movements) {
                if (pawn.isMoveValid(game.getBoard(), movement)) {
                    actions.add(new Action(pawn, movement));
                }
            }
        }

        Collections.shuffle(actions);
        return actions;
    }

    private static int evaluate(MovingBoard board) {
        return evaluate(board, board.getMaxPawns()) - evaluate(board, board.getMinPawns());
    }

    private static int evaluate(MovingBoard board, Collection<MovingPawn> pawns) {
        int score = pawns.size() * 5;

        for (MovingPawn pawn : pawns) {
            score += evaluate(board, pawn);
        }

        return score;
    }

    private static int evaluate(MovingBoard game, MovingPawn pawn) {
        if (pawn.getRow() == pawn.getPlayer().getGoal()) {
            return Integer.MAX_VALUE;
        }

        // Favorise les pièces qui peuvent gagner au prochain tour
        if (pawn.getRow() + pawn.getDirection() == pawn.getPlayer().getGoal()) {
            return Integer.MAX_VALUE / 2;
        }

        int score = 0;
        score += attackScore(game, pawn);

        // Favorise les attaques de notre côté
        if ((pawn.getPlayer().getGoal() - pawn.getRow()) / 2f < 4) {
            score *= 2;
        }

        score += getValidMoveCount(game, pawn) * 50;

        // Favorise les pushed
        if (!pawn.isPusher()) {
            score += 50;
        }

        // Favorise les pièces qui n'ont pas encore bougées
        if (pawn.getRow() == (pawn.getPlayer() == Player.RED ? 0 : 7)) {
            score += 50;
        }

        return score;
    }

    private static int attackScore(MovingBoard game, MovingPawn pawn) {
        int score = 0;
        if (nextToOtherPlayer(game, pawn, -1)) {
            score += 100;
        }
        if (nextToOtherPlayer(game, pawn, 1)) {
            score += 100;
        }
        return score;
    }

    private static boolean nextToOtherPlayer(MovingBoard game, MovingPawn pawn, int colOffset) {
        int row = pawn.getRow() + pawn.getDirection();
        int col = pawn.getCol() + colOffset;

        if (col < 0 || col > 7) {
            return false;
        }

        MovingPawn dest = game.getBoard()[row][col];
        return dest != null && dest.getPlayer() != pawn.getPlayer();
    }

    private static int getValidMoveCount(MovingBoard game, MovingPawn pawn) {
        int moveCount = 0;

        // Diagonale gauche
        if (pawn.getCol() > 0) {
            moveCount++;
        }

        // Il n'y a pas de pièce adverse devant
        if (!nextToOtherPlayer(game, pawn, 0)) {
            moveCount++;
        }

        // Diagonale droite
        if (pawn.getCol() < 7) {
            moveCount++;
        }

        return moveCount;
    }

    static class Action {
        private final MovingPawn pawn;
        private final Pawn.PawnMovement movement;

        public Action(MovingPawn pawn, Pawn.PawnMovement movement) {
            this.pawn = pawn;
            this.movement = movement;
        }

        public MovingPawn getPawn() {
            return pawn;
        }

        public Pawn.PawnMovement getMovement() {
            return movement;
        }
    }

    static class MiniMaxResult {
        private final int score;
        private final int row;
        private final int col;
        private final Pawn.PawnMovement movement;

        public MiniMaxResult(int score, int row, int col, Pawn.PawnMovement movement) {
            this.score = score;
            this.row = row;
            this.col = col;
            this.movement = movement;
        }

        public int getScore() {
            return score;
        }

        public int getRow() {
            return row;
        }

        public int getCol() {
            return col;
        }

        public Pawn.PawnMovement getMovement() {
            return movement;
        }

        @Override
        public String toString() {
            return "MiniMaxResult{" +
                    "score=" + score +
                    ", col=" + col +
                    ", row=" + row +
                    ", movement=" + movement +
                    '}';
        }
    }
}
